package serverP2P;
import java.util.Vector;
import java.io.File;
import java.io.IOException;
import java.net.InetAddress;
import java.net.SocketException;
import java.nio.file.Paths;
import java.nio.file.Files;
import java.net.ServerSocket;
import java.net.Socket;
import protocolP2P.ProtocolP2PPacketTCP;
import protocolP2P.ProtocolP2PPacket;
import protocolP2P.RequestResponseCode;
import protocolP2P.Payload;
import protocolP2P.LoadRequest;
import protocolP2P.FileList;
import protocolP2P.FilePart;
import exception.InternalError;
import exception.ProtocolError;
import exception.SizeError;
import exception.TransmissionError;
import exception.VersionError;
import remoteException.EmptyDirectory;
import remoteException.InternalRemoteError;
import remoteException.NotFound;
import remoteException.ProtocolRemoteError;
import remoteException.VersionRemoteError;
import remoteException.EmptyFile;
import java.util.Arrays;


/** Implementation of P2P-JAVA-PROJECT VERSION 1.0 protocol for TCP.
 * @author	Louis Royer
 * @author	Flavien Haas
 * @author	JS Auge
 * @version	1.0
 */
public class ServerManagementTCP implements Runnable {

	private String[] fileList;
	private String baseDirectory;
	private int TCPPort;
	private ServerSocket socket;

	/** Constructor for TCP implementation, with baseDirectory and TCPPort parameters.
	 * @param baseDirectory	the root directory where files are stored
	 * @param TCPPort	the server will listen on this port
	 */
	public ServerManagementTCP(String baseDirectory, int TCPPort) {
		this.baseDirectory = baseDirectory;
		this.TCPPort = TCPPort;
		initFileList();
		try {
			socket = new ServerSocket(TCPPort);
		} catch (SocketException e) {
			System.err.println("Error: cannot listen on port " + TCPPort + " (TCP)");
			System.exit(-1);
		} catch (IOException e) {
			System.err.println("Error: cannot openning TCP socket");
			System.exit(-2); 
		}
	}

	/** Implementation of runnable. This methods allows to run the server.
	 */
	public void run() {
		// TODO: handling multiple clients
		try {
			Socket s = socket.accept();
			System.err.println("Accepting new connection");
			while(true) {
				try {
					ProtocolP2PPacketTCP pd = new ProtocolP2PPacketTCP((Object)s);
					Payload p = pd.getPayload();
					switch (p.getRequestResponseCode()) {
						case LOAD_REQUEST:
							System.out.println("Received LOAD_REQUEST");
							assert p instanceof LoadRequest : "payload must be an instance of LoadRequest";
							if (!(p instanceof LoadRequest)) {
								sendInternalError(pd);
							} else {
								String filename = ((LoadRequest)p).getFilename();
								long offset = ((LoadRequest)p).getOffset();
								long maxSizePartialContent = ((LoadRequest)p).getMaxSizePartialContent();
								try {
									byte[] fullLoad = Files.readAllBytes(Paths.get(baseDirectory + filename));
									long sizeToSend = 0;
									if (fullLoad.length - offset < maxSizePartialContent) {
										System.out.println("Sending last partialContent");
										sizeToSend = fullLoad.length - offset;
									} else {
										sizeToSend = maxSizePartialContent;
									}
									System.out.println("maxSizePartialContent: " + maxSizePartialContent);
									System.out.println("Sending " + filename + " from " + offset + " to " + (offset + sizeToSend));
									byte[] load = Arrays.copyOfRange(fullLoad, (int)offset, (int)(offset + sizeToSend));
									if (Arrays.binarySearch(fileList, filename) >= 0) {
										try {
											if (load.length == 0) {
												pd.sendResponse((ProtocolP2PPacket)new ProtocolP2PPacketTCP(new Payload(RequestResponseCode.EMPTY_FILE)));
											} else {
												pd.sendResponse((ProtocolP2PPacket)new ProtocolP2PPacketTCP((Payload)(new FilePart(filename, fullLoad.length, offset, load))));
											}
										} catch (Exception e2) {
											System.err.println(e2);
										}
									} else {
										System.err.println("File requested not found: `" + filename + "` " + Arrays.binarySearch(fileList, filename));
										System.err.println("File list:");
										for (String f: fileList) {
											System.err.println("- " + f);
										}
										
										throw new IOException(); // to send a NOT_FOUND in the catch block
									}
								} catch (IOException e) {
									try {
										pd.sendResponse((ProtocolP2PPacket)new ProtocolP2PPacketTCP(new Payload(RequestResponseCode.NOT_FOUND)));
									} catch (Exception e2) {
										System.err.println(e2);
									}
								}
							}
							break;
						case LIST_REQUEST:
							System.out.println("Received LIST_REQUEST");
							try {
								if (fileList.length == 0) {
									System.err.println("Sending EMPTY_DIRECTORY");
									pd.sendResponse((ProtocolP2PPacket)new ProtocolP2PPacketTCP(new Payload(RequestResponseCode.EMPTY_DIRECTORY)));
								} else {
									System.out.println("Sending LIST_RESPONSE");
									pd.sendResponse((ProtocolP2PPacket)new ProtocolP2PPacketTCP((Payload)(new FileList(fileList))));
								}
							} catch (Exception e2) {
								System.err.println(e2);
							}
							break;
						default:
							sendInternalError(pd);
					}
				} catch (IOException e) {
				} catch (TransmissionError e) {
				} catch (ProtocolError e) {
				} catch (VersionError e) {
				} catch (InternalError e) {
				} catch (SizeError e) {
				}
			}
		} catch (IOException e) {
		}
	}

	/** Initialize local list of all files allowed to be shared.
	 */
	private void initFileList() {
		File folder = new File(baseDirectory);
		Vector<String> v = new Vector<String>();
		File[] files = folder.listFiles();
		/* Add non-recursively files's names to fileList */
		for (File f : files) {
			if (f.isFile()) {
				v.add(f.getName());
			}
		}
		fileList = new String[v.size()];
		v.toArray(fileList);
		Arrays.sort(fileList);
	}


	
	/** Send an internal error message.
	 * @param pd	ProtocolP2PPacketTCP to respond
	 */
	private void sendInternalError(ProtocolP2PPacketTCP pd) {
		try {
			pd.sendResponse((ProtocolP2PPacket)new ProtocolP2PPacketTCP(new Payload(RequestResponseCode.INTERNAL_ERROR)));
		} catch (Exception e) {
			System.err.println(e);
		}
	}

}

